print("import: 'AppKit'")
import AppKit

print("import: 'Cocoa'")
import Cocoa

print("import: 'CoreFoundation'")
import CoreFoundation

print("import: 'Foundation'")
import Foundation

print("import: 'PyObjCTools'")
import PyObjCTools

